/******************************************************************************
 *
 * Copyright:
 *    (C) 2006 Jere Knaappila
 *
 * File:
 *    poker.c
 *
 * Description:
 *    Implements Video Poker using "Jacks or Better"-rules
 *
 *****************************************************************************/




#include "poker.h"
#include "../pre_emptive_os/api/osapi.h"
#include "../pre_emptive_os/api/general.h"
#include "lcd.h"
#include "key.h"
#include "math.h"
#include "select.h"
#include <string.h>
#include <stdlib.h>

#define BIT(a) (1<<(a))

#define COLOR_BLACK 0x00
#define COLOR_WHITE 0xff
#define COLOR_RED   0xe0
#define COLOR_GREEN 0x1c
#define COLOR_BLUE  0x03

#define BACKGROUND_COLOR    COLOR_BLACK
#define DEFAULT_TEXT_COLOR  0xe0

#define SCREEN_HEIGHT 130
#define SCREEN_WIDTH  130

extern volatile tU32 ms;

enum poker_hand
{
    poker_hand_none,
    poker_hand_jacks_or_better,
    poker_hand_two_pairs,
    poker_hand_three_of_a_kind,
    poker_hand_straight,
    poker_hand_flush,
    poker_hand_full_house,
    poker_hand_four_of_a_kind,
    poker_hand_straight_flush,
    poker_hand_royal_flush,
    poker_hand_max,
};
char* poker_hand_names[]={
    "No Win",
    "Jacks or Better",
    "Two Pairs",
    "Three of a Kind",
    "Straight",
    "Flush",
    "Full House",
    "Four of a Kind",
    "Straight Flush"
    "Royal Flush"
};
int payout[6][poker_hand_max]={
    {0,0,0,0,0,0,0,0,0},
    {0,1,2,3,4,6,9,25,50,250},
    {0,2,4,6,8,12,18,50,100,500},    
    {0,3,6,9,12,18,27,75,150,750},        
    {0,4,8,12,16,24,36,100,200,1000},            
    {0,5,10,15,20,30,45,125,250,4000},                
};
enum suits
{
    suit_spade,
    suit_heart,
    suit_diamond,
    suit_club,
    suit_joker
};
unsigned char suit_symbols[4][9*9];
char * suit_symbols_char[]={
"\
    x    \
   xxx   \
  xxxxx  \
 xxxxxxx \
xxxxxxxxx\
xxxxxxxxx\
 xx x xx \
    x    \
   xxx   ",
"\
  x   x  \
 xxx xxx \
xxxxxxxxx\
xxxxxxxxx\
xxxxxxxxx\
 xxxxxxx \
  xxxxx  \
   xxx   \
    x    ",
"\
    x    \
   xxx   \
  xxxxx  \
 xxxxxxx \
xxxxxxxxx\
 xxxxxxx \
  xxxxx  \
   xxx   \
    x    ",
    
"\
    x    \
   xxx   \
   xxx   \
    x    \
 xx x xx \
xxxxxxxxx\
 xx x xx \
    x    \
   xxx   "

};
char* suit_ids="SHDCJ";
unsigned char suit_colors[]={COLOR_BLACK,COLOR_RED,COLOR_RED,COLOR_BLACK,COLOR_RED};
char* rank_ids="01234567890JQKA";

#define SUIT_OFFSET 		4
#define RANK_MASK 		((1<<SUIT_OFFSET)-1)
#define RANK_MAX		13
#define DECK_SIZE		(RANK_MAX*4)
#define GET_SUIT(card) 		(card>>SUIT_OFFSET)
#define GET_RANK(card) 		(card&RANK_MASK)
#define RANKS_EQUAL(a,b)	(GET_RANK(a)==GET_RANK(b))
#define MAKE_CARD(suit,rank) 	((suit<<SUIT_OFFSET)|rank)

#define CARD_H 35
#define CARD_W 20

static unsigned char player_hand[5];
static int player_credits;
static int player_bet;

struct deck
{
    unsigned char cards[DECK_SIZE];
    int size;
};
static struct deck playdeck;
/**
 * swap two cards
 */
inline void swap(unsigned char*a,unsigned char*b)
{
    unsigned char c=*a;
    *a=*b;
    *b=c;
}
/**
 * initialize 52 card deck with all cards, sorted
 */
void deck_initialize(struct deck*d)
{
    int suit;
    int rank;
    d->size=0;
    for(suit=0;suit<4;suit++)
    {
	for(rank=2;rank<15;rank++)
	{
	    d->cards[d->size++]=MAKE_CARD(suit,rank);
	}
    }
}
void deck_shuffle(struct deck*d)
{
    int t;
    for(t=0;t<d->size;t++)
    {
	int s=rand()%d->size;
	swap(&d->cards[t],&d->cards[s]);
    }
}
/**
 * gets topmost card from deck
 * 
 * @param d      deck to get card from
 * 
 * @return topmost card, or 0 if deck empty
 */
unsigned char deck_popcard(struct deck*d)
{
    if(d->size>0)
    {
	d->size--;
	return d->cards[d->size];
    }
    return 0;
}
void draw_suit_symbol(int x,int y,int suit)
{
    lcdIcon(x,y,9,9,0,0,suit_symbols[suit]);
}
/**
 * draw a card
 * 
 * @param card   card to draw
 * @param color  card border color
 */
void draw_card(int x,int y,unsigned char card,unsigned char color)
{
//    lcdRectBrd(x-1,y-1,CARD_W+2,CARD_H+2,COLOR_WHITE,color,color);
    lcdRectRound(x-1,y-1,CARD_W+2,CARD_H+2,COLOR_WHITE,color,COLOR_GREEN);
    lcdColor(COLOR_WHITE,suit_colors[GET_SUIT(card)]);
    draw_suit_symbol(x+2,y+2,GET_SUIT(card));
    lcdGotoxy(x+2,y+2+12);
    unsigned char rank=GET_RANK(card);
    if(rank==10)
    {
	lcdPutchar('1');
	lcdPutchar('0');
    }else{
	lcdPutchar(' ');
	lcdPutchar(rank_ids[rank]);
    }
	

}
/**
 * draw a decimal number
 */
void draw_num(int score)
{
    int t;
    unsigned char str[13];
    
    str[0] = score/100000 + '0';
    str[1] = (score/10000)%10 + '0';
    str[2] = (score/1000)%10 + '0';
    str[3] = (score/100)%10 + '0';
    str[4] = (score/10)%10 + '0';
    str[5] = score%10 + '0';
    str[6] = 0;
    for(t=0;t<5;t++)
    {
	if(str[t]=='0')
	    str[t]=' ';
	else
	    break;
    }
    
    lcdPuts(str);
}
void draw_score()
{
    lcdGotoxy(2,SCREEN_HEIGHT-13);
    lcdColor(COLOR_GREEN,COLOR_WHITE);
    lcdPuts("Credits:");
    draw_num(player_credits);
}
void draw_bet()
{
    lcdGotoxy(2,SCREEN_HEIGHT-26);
    lcdColor(COLOR_GREEN,COLOR_WHITE);
    lcdPuts("Bet:    ");
    draw_num(player_bet);
}
void draw_winnings(enum poker_hand w)
{
    lcdGotoxy(2,SCREEN_HEIGHT-40);
    lcdColor(COLOR_GREEN,COLOR_WHITE);
    lcdPuts(poker_hand_names[w]);
}
void draw_gameover()
{
    lcdGotoxy(10,SCREEN_HEIGHT-60);
    lcdColor(COLOR_GREEN,COLOR_WHITE);
    lcdPuts("Game Over");
}
void draw_cursor(int pos,unsigned char color)
{
    int x=5;
    int y=10;
    x+=(CARD_W+5)*pos+(CARD_W>>1);
    lcdGotoxy(x-3,y+CARD_H+5);
    lcdColor(COLOR_GREEN,color);
    lcdPutchar('^');
}
/**
 * draw hand 
 */
void draw_pokerhand(unsigned char *hand,unsigned char holdmask)
{
    int t;
    int x=5;
    int y=10;
    unsigned char border;

    for (t=0;t<5;t++) {
	
	if(holdmask&(1<<t))
	    border=COLOR_RED;
	else
	    border=COLOR_GREEN;
	    
	draw_card(x,y,hand[t],border);
	x+=CARD_W+5;

    }
}
/**
 * populate hand with top 5 cards from deck
 */
static void deck_deal(struct deck*d,unsigned char*hand)
{
    int t;        

    for(t=0;t<5;t++)
    {
	hand[t]=deck_popcard(&playdeck);
    }
}
/**
 * draw green background, bet and score
 * 
 */
static void draw_background()
{
    //clear screen
    lcdColor(COLOR_GREEN,COLOR_WHITE);
    lcdClrscr();
    draw_bet();
    draw_score();
}
/**
 * draw all game screen information
 * 
 * @param cursor cursor position, or -1 to disable
 */
static void draw_screen(int cursor)
{
    draw_background();
    draw_pokerhand(player_hand,0);
    if(cursor>=0)draw_cursor(cursor,COLOR_BLACK);
    draw_bet();
    draw_score();
}
/**
 * sort cards
 */
void cards_sort(unsigned char*cards,int size)
{
    int b,dirty;
    do
    {
	dirty=0;
	for (b=0;b<4;b++) {
	    if(GET_RANK(cards[b])>GET_RANK(cards[b+1]))
	    {
		swap(&cards[b],&cards[b+1]);
		dirty=1;
	    }
	}

    }while(dirty);
}

		
/**		 
 * finds winning hand		 
 * 		 
 * @param hand    cards in player hand, returns sorted		 
 * @param winmask returns cards involved in winning hand		 
 * 		 
 * @return winning hand		 
 */
enum poker_hand calculate_poker_hand(unsigned char *hand,unsigned char *winmask)
{
    int a,b;
    cards_sort(hand,5);
    
    //Straight flush
    *winmask=BIT(0)|BIT(1)|BIT(2)|BIT(3)|BIT(4);
    for(a=0;a<4;a++)
	if(GET_RANK(hand[a])!=(GET_RANK(hand[a+1])-1))break;
    if(a!=4)//if failed, check for straight flush ace low, lowest sorted card is 2
    {

	for(a=0;a<3;a++)
	    if(GET_RANK(hand[a])!=(GET_RANK(hand[a+1])-1))break;
	if(a==3)//check for ace, and lowest card
	{
	    if(GET_RANK(hand[0])==2 && GET_RANK(hand[4])==14)
		a=4;
	}
    }
    if(a==4)
    {
	//check straight flush
	for(a=0;a<4;a++)
	    if(GET_SUIT(hand[a])!=GET_SUIT(hand[a+1]))
		break;
	if(a==4)    
	{
	    if(GET_RANK(hand[4])==14)//ace high, royal flush
		return poker_hand_royal_flush;
	    return poker_hand_straight_flush;
	}
	 
	    
	return poker_hand_straight;
    }
    //four of a kind
    for(a=0;a<2;a++)
    {
	for (b=0;b<3;b++) {
	    if(!RANKS_EQUAL(hand[a+b],hand[a+b+1]))
		break;
	}
	if(b==3)
	{
	    *winmask=BIT(a)|BIT(a+1)|BIT(a+2)|BIT(a+3);
	    return poker_hand_four_of_a_kind;
	}

    }
    //full house (2+3)
    *winmask=BIT(0)|BIT(1)|BIT(2)|BIT(3)|BIT(4);
    if(RANKS_EQUAL(hand[0],hand[1]))
    {
	if(RANKS_EQUAL(hand[2],hand[3]) &&
	   RANKS_EQUAL(hand[3],hand[4]))
	    return poker_hand_full_house;
    }
    //full house (3+2)
    if(RANKS_EQUAL(hand[0],hand[1]) &&
       RANKS_EQUAL(hand[1],hand[2]))
    {
	if(RANKS_EQUAL(hand[3],hand[4]))
	    return poker_hand_full_house;
    }
    //flush
    for(a=0;a<4;a++)
	if(GET_SUIT(hand[a])!=GET_SUIT(hand[a+1]))
	    break;
    if(a==4)    
	return poker_hand_flush;
    //three of a kind
    for(a=0;a<3;a++)
    {
	for (b=0;b<2;b++) {
	    if(GET_RANK(hand[a+b])!=GET_RANK(hand[a+b+1]))
		break;
	}
	if(b==2)
	{
	    *winmask=BIT(a+0)|BIT(a+1)|BIT(a+2);
	    return poker_hand_three_of_a_kind;
	}

    }
    //two pairs
    // (2+2+1)
    if(RANKS_EQUAL(hand[0],hand[1]) &&
       RANKS_EQUAL(hand[2],hand[3]))
    {
	*winmask=BIT(0)|BIT(1)|BIT(2)|BIT(3);
	return poker_hand_two_pairs;
    }
    // (1+2+2)
    if(RANKS_EQUAL(hand[1],hand[2]) &&
       RANKS_EQUAL(hand[3],hand[4]))
    {
	*winmask=BIT(1)|BIT(2)|BIT(3)|BIT(4);
	return poker_hand_two_pairs;
    }
    // (2+1+2)
    if(RANKS_EQUAL(hand[0],hand[1]) &&
       RANKS_EQUAL(hand[3],hand[4]))
    {
	*winmask=BIT(0)|BIT(1)|BIT(3)|BIT(4);
	return poker_hand_two_pairs;
    }
    //one pair
    for(a=0;a<4;a++)
    {
	if(RANKS_EQUAL(hand[a],hand[a+1]) && (GET_RANK(hand[a])>=11))
	{
	    *winmask=BIT(a+0)|BIT(a+1);
	    return poker_hand_jacks_or_better;
	}
    }
    return poker_hand_none;
}
/**		 
 * finds winning hand		 
 * 		 
 * @param hand    cards in player hand
 * @param winmask returns cards involved in winning hand		 
 * 		 
 * @return winning hand	 
 */
enum poker_hand calculate_poker_hand_no_sort(unsigned char *hand,unsigned char *winmask)
{
    unsigned char thand[5];
    unsigned char twinmask=0;
    *winmask=0;
    memcpy(thand,hand,5);
    enum poker_hand retval=calculate_poker_hand(thand,&twinmask);
    //sort winmask back to original
    if(retval!=poker_hand_none)
    {
	int a,b;
	for (a=0;a<5;a++)
	{
	    if(twinmask&BIT(a))//if bit is set, search for matching card in original hand
	    {
		for(b=0;b<5;b++)
		{
		    if(thand[a]==hand[b])
			*winmask|=BIT(b);
		}
	    }
	}
    }
    return retval;
}
/**
 * initializes suit symbols
 */
void init_data()
{
    int t,s;
    for(s=0;s<4;s++)
    for(t=0;t<9*9;t++)
    {
	unsigned char c=suit_colors[s];
	char *a=suit_symbols_char[s]+t;

	switch(*a)
	{
	case 'x':
	    suit_symbols[s][t]=c;
	    break;
	case ' ':
	    suit_symbols[s][t]=COLOR_WHITE;
	    break;
	}
    }
}
/**
 * draw change bet menu
 * 
 */
void change_bet()
{
    tMenu menu;
	      
    menu.xPos = 10;
    menu.yPos = 5;
    menu.xLen = 6+(12*8);
    menu.yLen = 7*14+3;
    menu.noOfChoices = 6;
    menu.initialChoice = 0;
    menu.pHeaderText = "Change Bet";
    menu.headerTextXpos = 20;
    menu.pChoice[0] = "Bet 1 Credit";
    menu.pChoice[1] = "Bet 2 Credit";
    menu.pChoice[2] = "Bet 3 Credit";
    menu.pChoice[3] = "Bet 4 Credit";
    menu.pChoice[4] = "Bet 5 Credit";
    menu.pChoice[5] = "Return";
    menu.bgColor       = 0;
    menu.borderColor   = 0x6d;
    menu.headerColor   = 0;
    menu.choicesColor  = 0xfd;
    menu.selectedColor = 0xe0;
    if(player_credits<5)
    {
	menu.noOfChoices=player_credits+1;
	menu.pChoice[menu.noOfChoices-1]="Return";
    }
    int result=drawMenu(menu);
    if(result==(menu.noOfChoices-1))
	return;
    player_bet=result+1;

}
/**
 * draw main game menu
 * 
 * @return 1  - start game
 *         -1 - exit game
 *         0 - redraw screen
 */
int mainmenu()
{
    tMenu menu;

    menu.xPos = 10;
    menu.yPos = 20;
    menu.xLen = 6+(12*8);
    menu.yLen = 4*15;
    menu.noOfChoices = 3;
    menu.initialChoice = 0;
    menu.pHeaderText = "Video Poker";
    menu.headerTextXpos = 20;
    menu.pChoice[0] = "Deal";
    menu.pChoice[1] = "Change Bet";
    menu.pChoice[2] = "Exit";
    menu.bgColor       = 0;
    menu.borderColor   = 0x6d;
    menu.headerColor   = 0;
    menu.choicesColor  = 0xfd;
    menu.selectedColor = 0xe0;
    switch(drawMenu(menu))
    {
    case 0: return 1;
    case 1: change_bet();break;
    case 2: return -1;
    default: break;
    }
    return 0;
}
void playPoker()
{
  tU8 done = FALSE;
  tU8 anyKey;
  //game loop
  srand(ms);

  player_credits=10;
  player_bet=1;
  int t;
      int gamestate=0;
      unsigned char holdmask=0;
      unsigned char cursor=0;
      init_data();

  do
  {
      anyKey = checkKey();
      switch(gamestate)
      {
      case 0://main menu
	  draw_background();
	  draw_score();
	  draw_bet();
	  t=mainmenu();
	  if(t==1)
	      gamestate++;
	  if(t==-1)
	      return;

	  break;
      case 1://deal hand
	  deck_initialize(&playdeck);
	  deck_shuffle(&playdeck);
	  player_credits-=player_bet;
	  holdmask=0;
	  cursor=0;
	  deck_deal(&playdeck,player_hand);
	  gamestate++;
	  draw_screen(cursor);
      case 2://player picks cards to keep
	  {

	      switch(anyKey)
	      {
	      case KEY_LEFT:
		  draw_cursor(cursor,COLOR_GREEN);
		  if(cursor>0)cursor--;
		  draw_cursor(cursor,COLOR_BLACK);
		  break;
	      case KEY_RIGHT:
		  draw_cursor(cursor,COLOR_GREEN);
		  if(cursor<4)cursor++;
		  draw_cursor(cursor,COLOR_BLACK);
		  break;
	      case KEY_UP:
		  holdmask|=BIT(cursor);
		  draw_pokerhand(player_hand,holdmask);
		  break;
	      case KEY_DOWN:
		  holdmask&=~BIT(cursor);
		  draw_pokerhand(player_hand,holdmask);
		  break;
	      case KEY_CENTER:
		  gamestate++;
		  break;
	      }
	  }
	  break;
      case 3://deal new cards for picked slots
	  for(t=0;t<5;t++)
	  {
	      if(!(holdmask&BIT(t)))
	      {
		  player_hand[t]=deck_popcard(&playdeck);
	      }
	  }
	  gamestate++;
	  break;
      case 4:
	  {
	      unsigned char winmask;
	      enum poker_hand win=calculate_poker_hand_no_sort(player_hand,&winmask);
	      player_credits+=payout[player_bet][win];
	      if(player_bet>player_credits)player_bet=player_credits;
	      draw_winnings(win);
	      draw_pokerhand(player_hand,winmask);
	      draw_bet();
	      draw_score();

	      if(player_credits==0)
	      {
		  draw_gameover();
		  gamestate=10;
	      }else{
		  gamestate++;
	      }
	  }

	  break;
      case 5:
	  if(anyKey!=KEY_NOTHING)
	      gamestate=0;
	  break;
      case 10://Game over state
	  if(anyKey!=KEY_NOTHING)
	  {
	      return;
	  }
	  break;
      }
      osSleep(10);
  }while(done==FALSE);
}
